import os

import numpy as np
from bw2data import geomapping, projects
from bw2data.data_store import ProcessedDataStore

from .meta import loadings
from .utils import get_pandarus_map, create_certain_datapackage
from .validate import loading_validator


class Loading(ProcessedDataStore):
    """ """

    _metadata = loadings
    validator = loading_validator
    matrix = "loading_matrix"

    def add_geomappings(self, data):
        """In theory, this shouldn't do anything, as all spatial units should be in defined by the method."""
        geomapping.add({obj[1] for obj in data})

    def process(self, **extra_metadata):
        data = self.load()
        create_certain_datapackage([(geomapping[line[1]], 0) for line in data], [line[0] for line in data], self, **extra_metadata)

    @property
    def filename(self):
        return super(Loading, self).filename + ".loading"

    def write_to_map(self, method, flow, geocollection=None):
        map_obj = get_pandarus_map(method, geocollection)
        data = {x[1][1]: x[0] for x in self.load()}
        if map_obj.raster:
            self._write_raster_loadings_to_map(map_obj, data)
        else:
            self._write_vector_loadings_to_map(map_obj, data)

    def _write_raster_loadings_to_map(self, map_obj, data):
        NODATA = -9999.0
        filepath = os.path.join(projects.output_dir, self.filename + ".tiff")

        array = np.zeros(map_obj.file.array().shape) + NODATA
        for obj in map_obj:
            if obj["label"] in data:
                array[obj["row"], obj["col"]] = data[obj["label"]]

        map_obj.file.write_modified_array(filepath, array, nodata=NODATA)
        return filepath

    def _write_vector_loadings_to_map(self, map_obj, data):
        raise NotImplementedError



import wrapt
from bw2data import config, projects
from bw2data.tests import BW2DataTest

from bw2regional.meta import geocollections, intersections, loadings


class BW2RegionalTest(BW2DataTest):
    def extra_setup(self):
        geocollections.__init__()
        intersections.__init__()
        loadings.__init__()


@wrapt.decorator
def bw2test(wrapped, instance, args, kwargs):
    config.dont_warn = True
    config.is_test = True
    config.cache = {}
    projects._use_temp_directory()
    projects.set_current("default")
    geocollections.__init__()
    intersections.__init__()
    loadings.__init__()
    return wrapped(*args, **kwargs)



from .loading import Loading
from .meta import extension_tables, geocollections
from .utils import get_pandarus_map
from .validate import xtable_validator


class ExtensionTable(Loading):
    _metadata = extension_tables
    validator = xtable_validator
    matrix = "xtable_matrix"

    @property
    def filename(self):
        return super(ExtensionTable, self).filename.replace(".loading", ".xtable")

    def write_to_map(self, *args, **kwargs):
        raise NotImplementedError

    def import_from_map(self):
        geocollection = extension_tables[self.name].get("geocollection")
        xt_field = extension_tables[self.name].get("xt_field")
        # TODO: Handle raster band
        # band = extension_tables[self.name].get("band")

        if not geocollection:
            raise ValueError("No geocollection for this extension table")

        map_obj = get_pandarus_map(geocollection)
        data = []

        if map_obj.vector:
            if xt_field is None:
                raise ValueError("No `xt_field` field name specified")

            id_field = geocollections[geocollection].get("field")
            if not id_field:
                raise ValueError(
                    "Geocollection must specify ``field`` field name for unique feature ids"
                )

        for feature in map_obj:
            if map_obj.vector:
                label = feature["properties"][id_field]
                value = float(feature["properties"][xt_field])
            else:
                label = feature["label"]
                value = feature["value"]
            data.append((value, (geocollection, label)))

        self.write(data)



import os
import pickle

import numpy as np
import pandas as pd
from bw2data import JsonWrapper, geomapping, projects

from . import (
    ExtensionTable,
    Intersection,
    Topography,
    geocollections,
    intersections,
    topocollections,
)
from .utils import create_certain_datapackage

from bw_processing import INDICES_DTYPE


def relabel(data, first, second):
    """Add geocollection names to geo identifiers"""
    return [((first, a), (second, b), c) for a, b, c in data]


def load_file(filepath):
    """Load Pandarus JSON output file.

    Returns metadata and calculation results."""
    try:
        obj = JsonWrapper.load_bz2(filepath)
    except:
        obj = JsonWrapper.load(filepath)
    return obj["metadata"], obj["data"]


def get_possible_collections(kwargs):
    """Return all geo- and topocollections for a file hash.

    Returns list of (collection name, collection type) tuples."""
    candidates = {"sha256", "field", "layer", "band"}
    filters = {k: v for k, v in kwargs.items() if k in candidates and kwargs[k]}

    def match(other):
        return all(other.get(key) == filters[key] for key in filters)

    possibles = {
        (name, "geocollection")
        for name, value in geocollections.items()
        if match(value)
    }.union(
        {
            (name, "topocollection")
            for name, value in topocollections.items()
            if match(value) and not value["empty"]
        }
    )
    return possibles


def import_from_pandarus(fp):
    """Load output file from Pandarus job.

    This function will:

    * Load pandarus output file
    * Locate the appropriate geo- or topocollection
    * Check to make sure that SHA256 hashes and other metadata match
    * If ``first`` is a topocollection, make sure the appropriate ``Topology`` exists, and squash the pandarus results to the linked geocollection(s).

    """
    assert os.path.isfile(fp)
    metadata, data = load_file(fp)

    # Check metadata
    assert "first" in metadata and "second" in metadata, "Invalid metadata in file"

    first_collections = get_possible_collections(metadata["first"])
    second_collections = get_possible_collections(metadata["second"])

    # No self-intersections
    assert not first_collections.intersection(
        second_collections
    ), "Overlapping geocollections"

    if "topocollection" in {x[1] for x in second_collections}.union(
        {x[1] for x in first_collections}
    ):
        return handle_topographical_intersection(
            metadata, data, first_collections, second_collections, fp
        )

    # Only one geocollection in both `first` and `second`
    assert len(first_collections) == 1, "Must intersect with exactly one geocollection"
    assert len(second_collections) == 1, "Must intersect with exactly one geocollection"
    first = first_collections.pop()[0]
    second = second_collections.pop()[0]

    assert (
        first,
        second,
    ) not in intersections, "Intersection between {} and {} already exists".format(
        first, second
    )

    dataset = relabel(data, first, second)
    intersection = Intersection((first, second))
    intersection.register(filepath=fp)
    intersection.write(dataset)
    intersection.create_reversed_intersection()

    return first, second


def handle_topographical_intersection(
    metadata, data, first_collections, second_collections, filepath
):
    """Handle an intersection between one or more topographies and a single geocollection.

    Each topography is associated with exactly one geocollection.

    Each topography is not empty, i.e. we can use the topographical definitions to filter.

    The procedure is:
    #. Check metadata validity, and make sure the topography ids are in the first column
    #. To split data into each topography
    #. Squash the topography to geocollections
    #. Create a new intersection for each geocollection/topography pair

    We use Pandas DataFrames to do aggregation in a resource efficient way. We also write the processed Intersection arrays directly.

    """
    # Check that topography(s) are in either first or second position, and
    # switch to first position if necessary
    first_labels = {x[1] for x in first_collections}
    second_labels = {x[1] for x in second_collections}
    if first_labels == {"topocollection"}:
        assert (
            len(second_collections) == 1
        ), "Must intersect with exactly one geocollection"
        assert second_labels == {
            "geocollection"
        }, "Must intersect topography with geocollections"
        lbl = list(second_collections)[0][0]
        geomapping.add({(lbl, y) for x, y, z in data})
        data = pd.DataFrame(
            [(x, geomapping[(lbl, y)], z) for x, y, z in data],
            columns=["topo_id", "feature_mapped_id", "area"],
        )
    elif second_labels == {"topocollection"}:
        assert (
            len(first_collections) == 1
        ), "Must intersect with exactly one geocollection"
        assert first_labels == {
            "geocollection"
        }, "Must intersect topography with geocollections"
        lbl = list(first_collections)[0][0]
        geomapping.add({(lbl, x) for x, y, z in data})
        data = pd.DataFrame(
            [(y, geomapping[(lbl, x)], z) for x, y, z in data],
            columns=["topo_id", "feature_mapped_id", "area"],
        )
        metadata["first"], metadata["second"] = metadata["second"], metadata["first"]
        first_collections, second_collections = second_collections, first_collections
    else:
        raise ValueError(
            "Intersections between mixed topocollections and "
            "geocollections are not supported"
        )

    grouped_data = data.groupby(["topo_id"])
    topo_geocollections = [
        topocollections[name]["geocollection"] for name, kind in first_collections
    ]
    topo_data = [
        {geomapping[k]: set(v) for k, v in Topography(obj[0]).load().items()}
        for obj, name in zip(first_collections, topo_geocollections)
    ]
    other_geocollection = second_collections.pop()[0]

    for name in topo_geocollections:
        assert (
            other_geocollection,
            name,
        ) not in intersections, "Intersection between {} and {} already exists".format(
            other_geocollection, name
        )

    for name, mapping in zip(topo_geocollections, topo_data):
        print("Merging topographical faces for geocollection {}".format(name))
        indices_arrays, data_arrays = [], []
        valid_topo_ids = data["topo_id"].unique()

        for key in mapping:
            try:
                # Get list of dataframes, one per face id, all
                # from a given region
                all_intersections_for_a_region = [
                    grouped_data.get_group(topo_id)
                    for topo_id in mapping[key]
                    if topo_id in valid_topo_ids
                ]
            except KeyError:
                continue

            if not all_intersections_for_a_region:
                continue

            temp = (
                pd.concat(all_intersections_for_a_region)
                .groupby("feature_mapped_id")
                .sum()
            )
            index_array = np.empty(len(temp), dtype=INDICES_DTYPE)
            index_array["row"] = key
            index_array["col"] = temp.index.values
            indices_arrays.append(index_array)
            data_arrays.append(np.ndarray(temp["area"].values))

        assert indices_arrays, "Empty intersection"

        indices_arrays = np.hstack(indices_arrays)
        data_arrays = np.hstack(data_arrays)

        print("Creating intersection ({}, {})".format(name, other_geocollection))
        intersection = Intersection((name, other_geocollection))
        intersection.register(filepath=filepath)
        filepath = os.path.join(
            projects.dir, "processed", intersection.filename + ".npy"
        )
        with open(filepath, "wb") as f:
            pickle.dump(arrays, f, protocol=pickle.HIGHEST_PROTOCOL)

        # TODO: Flip row and column?

        intersection = Intersection((other_geocollection, name))
        intersection.register(filepath=filepath)
        filepath = os.path.join(
            projects.dir, "processed", intersection.filename + ".npy"
        )
        with open(filepath, "wb") as f:
            pickle.dump(arrays, f, protocol=pickle.HIGHEST_PROTOCOL)

    return [(n, other_geocollection) for n in topo_geocollections]


def import_xt_from_rasterstats(fp, name, gc, **kwargs):
    metadata, data = load_file(fp)
    assert "vector" in metadata and "raster" in metadata, "Invalid metadata in file"

    vector = get_possible_collections(metadata["vector"])
    assert len(vector) == 1, "Must intersect with exactly one geocollection"

    vector = list(vector)[0]

    assert vector[1] != "topocollection"

    xt = ExtensionTable(name)
    md = {
        "filepath": fp,
        "vector": metadata["vector"],
        "raster": metadata["raster"],
        "geocollection": gc,
    }
    md.update(**kwargs)
    xt.register(**md)
    xt.write(
        [
            (row[1]["mean"], (vector[0], row[0]))
            for row in data
            if row[1]["mean"] is not None
        ]
    )
    return xt



import hashlib


def sha256(filepath, blocksize=65536):
    """Generate SHA 256 hash for file at `filepath`"""
    hasher = hashlib.sha256()
    fo = open(filepath, "rb")
    buf = fo.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fo.read(blocksize)
    return hasher.hexdigest()



__all__ = (
    "cg",
    "create_ecoinvent_collections",
    "create_empty_intersection",
    "create_restofworlds_collections",
    "create_world_collections",
    "divide_by_area",
    "extension_tables",
    "ExtensionTable",
    "ExtensionTablesLCA",
    "geocollections",
    "get_spatial_dataset_kind",
    "hash_collection",
    "import_from_pandarus",
    "import_regionalized_cfs",
    "Intersection",
    "intersections",
    "label_activity_geocollections",
    "Loading",
    "loadings",
    "OneSpatialScaleLCA",
    "PandarusRemote",
    "remote",
    "reset_all_geo",
    "reset_geo_meta",
    "sha256",
    "topocollections",
    "Topography",
    "TwoSpatialScalesLCA",
    "TwoSpatialScalesWithGenericLoadingLCA",
)

from .version import version as __version__

from constructive_geometries import ConstructiveGeometries

cg = ConstructiveGeometries()

from bw2data import config

from .topography import Topography
from .intersection import Intersection
from .loading import Loading
from .meta import (
    extension_tables,
    geocollections,
    intersections,
    loadings,
    topocollections,
)
from .xtables import ExtensionTable

from .databases import label_activity_geocollections
from .density import divide_by_area
from .hashing import sha256
from .lca import (
    ExtensionTablesLCA,
    OneSpatialScaleLCA,
    TwoSpatialScalesLCA,
    TwoSpatialScalesWithGenericLoadingLCA,
)
from .pandarus import import_from_pandarus
from .pandarus_remote import PandarusRemote, remote
from .utils import (
    create_empty_intersection,
    get_spatial_dataset_kind,
    hash_collection,
    import_regionalized_cfs,
    reset_all_geo,
    reset_geo_meta,
)
from .base_data import create_ecoinvent_collections, create_world_collections, create_restofworlds_collections

config.metadata.extend(
    [extension_tables, geocollections, topocollections, intersections, loadings]
)



from pathlib import Path

from bw2data.serialization import CompoundJSONDict, SerializedDict

from .hashing import sha256


class Loadings(SerializedDict):
    """Metadata on regionalized LCIA weightings."""

    filename = "loadings.json"


class Intersections(CompoundJSONDict):
    """Areal intersections between the elements of two geo- or topocollections"""

    filename = "intersections.json"


class Geocollections(SerializedDict):
    """Metadata for spatial data sets."""

    filename = "geocollections.json"

    def __setitem__(self, key, value):
        from .utils import get_spatial_dataset_kind

        if "filepath" in value:
            assert Path(value["filepath"]).is_file(), "Can't find file at `{}`".format(
                value["filepath"]
            )
            value["sha256"] = sha256(value["filepath"])
            value["kind"] = get_spatial_dataset_kind(value["filepath"])
        super(Geocollections, self).__setitem__(key, value)


class Topocollections(Geocollections):
    """Mappings from geocollections to a set of topographical face ids."""

    filename = "topocollections.json"

    def __setitem__(self, key, value):
        value["empty"] = value.get("empty", True)
        if "geocollection" not in value:
            raise ValueError(
                "topocollections must be linked to exactly one geocollection"
            )
        elif value["geocollection"] not in geocollections:
            raise ValueError(
                "Linked geocollection {} does not exist".format(value["geocollection"])
            )
        super(Topocollections, self).__setitem__(key, value)


class ExtensionTables(SerializedDict):
    """Metadata for extension tables that give loadings on a third spatial scale."""

    filename = "extension-tables.json"


extension_tables = ExtensionTables()
geocollections = Geocollections()
intersections = Intersections()
loadings = Loadings()
topocollections = Topocollections()



import os
import time
import uuid

import requests
import wrapt
from bw2data import config, projects

from . import geocollections, intersections, topocollections
from .errors import WindowsPathCharacterLimit
from .pandarus import import_from_pandarus, import_xt_from_rasterstats
from .utils import hash_collection


class RemoteError(Exception):
    """Can't reach pandarus-remote web service"""

    pass


class NotYetCalculated(Exception):
    """Resource hasn't been calculated yet"""

    pass


class AlreadyExists(Exception):
    """Resource has already been calculated"""

    pass


class PendingJob(object):
    """A calculation job enqueued on a remote server"""

    def __init__(self, url):
        self.url = url

    @property
    def status(self):
        response = requests.get(self.url)
        if response.status_code != 404:
            return response.text
        else:
            return "forgotten"

    def poll(self, interval=10):
        try:
            while True:
                if self.status not in {"failed", "finished", "forgotten"}:
                    time.sleep(interval)
                else:
                    break
        except KeyboardInterrupt:
            pass
        finally:
            print("\nJob ended with status '{}'".format(self.status))


@wrapt.decorator
def check_alive(wrapped, instance, args, kwargs):
    if not instance.alive:
        raise RemoteError("Can't reach {}".format(instance.url))
    return wrapped(*args, **kwargs)


class PandarusRemote(object):
    """Interaction with `pandarus_remote <https://github.com/cmutel/pandarus_remote>`__ web service.

    Default URL is `https://pandarus.brightway.dev`."""

    def __init__(self, url=None):
        self.url = url or "https://pandarus.brightway.dev"
        if self.url[-1] == "/":
            self.url = self.url[:-1]

    @property
    def alive(self):
        return requests.get(self.url).status_code == 200

    def _download_file(self, resp):
        assert "Content-Disposition" in resp.headers
        download_dirpath = projects.request_directory("regional")

        filepath = os.path.abspath(
            os.path.join(
                download_dirpath,
                resp.headers["Content-Disposition"].replace(
                    "attachment; filename=", ""
                ),
            )
        )

        if config._windows and len(str(filepath)) > 250:
            # Windows has an absolute limit of 255 characters in a filepath
            if len(str(os.path.abspath(download_dirpath))) > 200:
                ERROR = """Cannot safely save files in this project directory.
                The project name is too long: {} characters for complete directory path, should fewer than 200.
                The directory used for downloads is: {}
                Please start a new project with a shorter project name."""
                raise WindowsPathCharacterLimit(
                    ERROR.format(len(download_dirpath), download_dirpath)
                )
            filepath = os.path.abspath(
                os.path.join(
                    download_dirpath,
                    uuid.uuid4().hex + filepath.split(".")[-1],
                )
            )

        chunk = 128 * 1024
        with open(filepath, "wb") as f:
            while True:
                segment = resp.raw.read(chunk)
                if not segment:
                    break
                f.write(segment)
        return filepath

    @check_alive
    def catalog(self):
        return requests.get(self.url + "/catalog").json()

    @check_alive
    def status(self, url):
        return requests.get(self.url + url).text

    @check_alive
    def upload(self, collection):
        if collection in topocollections:
            metadata = topocollections[collection]
        elif collection in geocollections:
            metadata = geocollections[collection]
        else:
            raise ValueError("Unknown geocollection {}".format(collection))

        assert "filepath" in metadata, "Can't find file for this collection"

        try:
            collection_hash = metadata["sha256"]
        except KeyError:
            collection_hash = hash_collection(collection)

        if collection_hash in {obj[1] for obj in self.catalog()["files"]}:
            raise AlreadyExists

        url = self.url + "/upload"
        data = {
            "layer": metadata.get("layer") or "",
            "field": metadata.get("field") or "",
            "band": metadata.get("band") or "",
            "sha256": collection_hash,
            "name": os.path.basename(metadata["filepath"]),
        }
        files = {"file": open(metadata["filepath"], "rb")}
        resp = requests.post(url, data=data, files=files)
        if resp.status_code == 200:
            return resp.json()
        else:
            raise RemoteError("{}: {}".format(resp.status_code, resp.text))

    @check_alive
    def intersection(self, collection_one, collection_two):
        if (collection_one, collection_two) in intersections:
            print(
                "Skipping existing intersection: ({}, {})".format(
                    collection_one, collection_two
                )
            )
            return

        catalog = self.catalog()
        first = self.hash_and_upload(collection_one, catalog)
        second = self.hash_and_upload(collection_two, catalog)

        resp = requests.post(
            self.url + "/intersection",
            data={"first": first, "second": second},
            stream=True,
        )
        if resp.status_code == 404:
            raise NotYetCalculated(
                "Not yet calculated; Run `.calculate_intersection` first."
            )
        self.handle_errors(resp)

        filepath = self._download_file(resp)
        return import_from_pandarus(filepath)

    @check_alive
    def intersection_as_new_geocollection(
        self, collection_one, collection_two, new_name
    ):
        if new_name in geocollections:
            print("Skipping creation of existing geocollection")
            return

        catalog = self.catalog()
        first = self.hash_and_upload(collection_one, catalog)
        second = self.hash_and_upload(collection_two, catalog)

        resp = requests.post(
            self.url + "/intersection-file",
            data={"first": first, "second": second},
            stream=True,
        )
        if resp.status_code == 404:
            raise NotYetCalculated(
                "Not yet calculated; Run `.calculate_intersection` first."
            )
        self.handle_errors(resp)

        filepath = self._download_file(resp)

        geocollections[new_name] = {
            "filepath": filepath,
            "field": "id",
            "url": self.url + "/intersection-file",
            "is intersection": True,
            "first": collection_one,
            "second": collection_two,
        }

        try:
            self.intersection(new_name, collection_one)
        except NotYetCalculated:
            self.calculate_intersection(new_name, collection_one)
            print(
                """Remote is calculating intersection, run the following when done:
                  remote.intersection("{}", "{}"")""".format(
                    new_name, collection_one
                )
            )
        try:
            self.intersection(new_name, collection_two)
        except NotYetCalculated:
            self.calculate_intersection(new_name, collection_two)
            print(
                """Remote is calculating intersection, run the following when done:
                  remote.intersection("{}", "{}"")""".format(
                    new_name, collection_two
                )
            )

    @check_alive
    def rasterstats_as_xt(self, vector, raster, name):
        """ """
        catalog = self.catalog()
        first = self.hash_and_upload(vector, catalog)
        second = self.hash_and_upload(raster, catalog)

        resp = requests.post(
            self.url + "/rasterstats",
            data={"vector": first, "raster": second},
            stream=True,
        )
        self.handle_errors(resp)

        filepath = self._download_file(resp)
        return import_xt_from_rasterstats(filepath, name, vector)

    @check_alive
    def calculate_rasterstats(self, vector, raster):
        catalog = self.catalog()
        first = self.hash_and_upload(vector, catalog)
        second = self.hash_and_upload(raster, catalog)

        resp = requests.post(
            self.url + "/calculate-rasterstats",
            data={"vector": first, "raster": second},
        )
        self.handle_errors(resp)

        print("Calculation job submitted.")
        return PendingJob(self.url + resp.text)

    @check_alive
    def calculate_intersection(self, collection_one, collection_two):
        catalog = self.catalog()
        first = self.hash_and_upload(collection_one, catalog)
        second = self.hash_and_upload(collection_two, catalog)

        resp = requests.post(
            self.url + "/calculate-intersection",
            data={"first": first, "second": second},
        )
        self.handle_errors(resp)

        print("Calculation job submitted.")
        return PendingJob(self.url + resp.text)

    def hash_and_upload(self, collection, catalog=None):
        hashes = {obj[1] for obj in (catalog or self.catalog())["files"]}
        hashed = hash_collection(collection)
        if not hashed:
            raise ValueError("Can't find collection {}".format(collection))
        if hashed not in hashes:
            self.upload(collection)
        return hashed

    def handle_errors(self, response):
        if response.status_code == 409:
            raise AlreadyExists
        elif response.status_code != 200:
            raise ValueError(
                "Server returned an error code: {}: {}".format(
                    response.status_code, response.text
                )
            )


remote = PandarusRemote()



import copy
import os
import shutil
from bw_processing import load_datapackage
from fs.zipfs import ZipFS

import fiona
import numpy as np
import rasterio
from bw2data import Method, methods, projects
from scipy import sparse

from .errors import MissingSpatialSourceData, SiteGenericMethod
from .hashing import sha256
from .meta import (
    extension_tables,
    geocollections,
    intersections,
    loadings,
    topocollections,
)

from bw_processing import INDICES_DTYPE, clean_datapackage_name, create_datapackage
from fs.zipfs import ZipFS


def filter_fiona_metadata(dct):
    """Include only valid Fiona keywords for opening a feature collection"""
    valid_keys = {"crs", "driver", "schema", "layer", "vfs"}
    return {k: v for k, v in dct.items() if k in valid_keys}


def import_regionalized_cfs(
    geocollection, method, mapping, overwrite=True, global_cfs=None, scaling_factor=1
):
    """Import data from a geospatial dataset (i.e. raster or vector data) into a ``Method``.

    A ``Method`` can have both site-generic and regionalized characterization factors.

    The ``mapping`` defines which field (vector) or band (raster) maps to which biosphere flows. Some geocollections may only define regionalized chracterization factors for a single biosphere flow, but it is much more common to have each field or band map to multiple biosphere flows. Therefore, mapping should be defined as:

    .. code-block:: python

        {
            field name (str) or raster band (int): [list of biosphere flows (tuples)]
        }

    Args:
        * *geocollection*: A ``geocollection`` name.
        * *method*: A ``Method`` instance (not the identifying tuple).
        * *mapping*: Mapping from fields or bands to biosphere flows. See above.
        * *overwrite*: Boolean. Overwrite existing characterization factors. Default is ``True``. Set to ``False`` if ``method`` already has site-generic characterization factors.
        * *global_cfs*: An optional list of CFs to add when writing the method.
        * *scaling_factor*: Optional. Rescale the values in the spatial data source.

    """
    try:
        from pandarus import Map
    except:
        raise ImportError("`pandarus` is required for this function")
    if not isinstance(method, Method):
        raise TypeError(
            "Must pass bw2data Method instance (got %s: %s" % (type(method), method)
        )
    assert geocollection in geocollections

    if geocollections[geocollection]["kind"] != "vector":
        raise ValueError("Geocollection must be a vector")

    if global_cfs is None:
        global_cfs = []

    metadata = copy.deepcopy(method.metadata)
    metadata.update(geocollections[geocollection])

    if overwrite:
        data = []
    else:
        data = method.load()

    map_obj = Map(metadata.pop("filepath"), **filter_fiona_metadata(metadata))

    id_field = geocollections[geocollection].get("field")
    if not id_field:
        raise ValueError(
            "Geocollection must specify ``field`` field name for unique feature ids"
        )

    for feature in map_obj:
        for field in mapping:
            for flow in mapping[field]:
                data.append(
                    (
                        flow,  # Biosphere flow
                        float(feature["properties"][field])
                        * scaling_factor,  # CF value
                        (
                            geocollection,
                            feature["properties"][id_field],
                        ),  # Spatial unit
                    )
                )

    method.write(data + global_cfs)

    if overwrite:
        methods[method.name]["geocollections"] = [geocollection]
        methods.flush()


def get_pandarus_map(geocollection):
    try:
        from pandarus import Map
    except:
        raise ImportError("`pandarus` is required for this function")
    if geocollection not in geocollections:
        raise ValueError("Geocollection %s not registered" % geocollection)
    geocollection = geocollections[geocollection]
    if not geocollection.get("filepath"):
        raise MissingSpatialSourceData("No filepath given for geocollection")
    metadata = {
        k: v for k, v in geocollection.items() if v is not None and k != "filepath"
    }
    return Map(geocollection["filepath"], **metadata)


def get_pandarus_map_for_method(method, geocollection=None):
    try:
        from pandarus import Map
    except:
        raise ImportError("`pandarus` is required for this function")
    if not methods[method].get("geocollections", []):
        raise SiteGenericMethod
    elif len(methods[method]["geocollections"]) > 1 and geocollection is None:
        raise ValueError("Must specify geocollection for this LCIA method")
    assert method in methods, "Unknown LCIA method"
    method_data = methods[method]
    geocollection = geocollections[
        geocollection or methods[method]["geocollections"][0]
    ]
    if not geocollection.get("filepath"):
        raise MissingSpatialSourceData("No filepath given for geocollection")
    metadata = {
        "band": method_data.get("band"),
        "layer": geocollection.get("layer"),
        "field": geocollection.get("field"),
        "vfs": geocollection.get("vfs"),
        "encoding": geocollection.get("encoding"),
    }
    metadata = {k: v for k, v in metadata.items() if v is not None}
    return Map(geocollection["filepath"], **metadata)


def hash_collection(name):
    """Return SHA256 hash for a topo- or geocollection.

    Prefers topocollection if available.
    """
    if name in topocollections:
        if "sha256" in topocollections[name]:
            return topocollections[name]["sha256"]
        try:
            assert os.path.isfile(topocollections[name]["filepath"])
            return sha256(topocollections[name]["filepath"])
        except KeyError:
            pass
    if name in geocollections:
        if "sha256" in geocollections[name]:
            return geocollections[name]["sha256"]
        try:
            assert os.path.isfile(geocollections[name]["filepath"])
            return sha256(geocollections[name]["filepath"])
        except KeyError:
            pass
    return False


def create_empty_intersection(name):
    """Shortcut to create Intersection object with no data"""
    from .intersection import Intersection
    inter = Intersection(name)
    inter.register()
    inter.write([])
    return inter


def get_spatial_dataset_kind(filepath):
    """Get kind of spatial dataset at `filepath`.

    Returns one of "vector", "raster", None.

    """
    with fiona.Env():
        try:
            with fiona.open(filepath) as source:
                assert source.meta
                return "vector"
        except:
            pass
    with rasterio.Env():
        try:
            with rasterio.open(filepath) as source:
                assert source.meta
                return "raster"
        except:
            pass
    return None


def reset_all_geo():
    """Reset all bw2regional data and metadata"""
    all_meta = (
        extension_tables,
        geocollections,
        intersections,
        loadings,
        topocollections,
    )
    shutil.rmtree(projects.request_directory("regional"))
    projects.request_directory("regional")
    for meta in all_meta:
        meta.data = {}
        meta.flush()


def reset_geo_meta():
    intersections.__init__()
    loadings.__init__()
    geocollections.__init__()
    topocollections.__init__()
    extension_tables.__init__()


def filter_rows(matrix, row_indices, exclude=True):
    """Filter a sparse matrix, either excluding or taking only the rows in ``row_indices``.

    * ``matrix``: A Scipy sparse matrix.
    * ``row_indices``: An iterable of integer row indices
    * ``exclude``: Boolean. If true, exclude rows in ``row_indices``. Otherwise, include only rows in ``row_indices``.

    Returns a sparse matrix.
    """
    matrix = matrix.tocoo()
    indices = np.array(row_indices)
    row_mask = np.in1d(matrix.row, indices, invert=exclude)
    return sparse.coo_matrix(
        (matrix.data[row_mask], (matrix.row[row_mask], matrix.col[row_mask])),
        matrix.shape,
    ).tocsr()


def filter_columns(matrix, col_indices, exclude=True):
    """Filter a sparse matrix, either excluding or taking only the columns in ``col_indices``.

    * ``matrix``: A Scipy sparse matrix.
    * ``col_indices``: An iterable of integer column indices
    * ``exclude``: Boolean. If true, exclude rows in ``row_indices``. Otherwise, include only rows in ``row_indices``.

    Returns a sparse matrix.
    """
    matrix = matrix.tocoo()
    indices = np.array(col_indices)
    col_mask = np.in1d(matrix.col, indices, invert=exclude)
    return sparse.coo_matrix(
        (matrix.data[col_mask], (matrix.row[col_mask], matrix.col[col_mask])),
        matrix.shape,
    ).tocsr()


def create_certain_datapackage(indices, data, data_store, **extra_metadata):
    data_array = np.array(data)
    indices_array = np.array(indices, dtype=INDICES_DTYPE)

    dp = create_datapackage(
        fs=ZipFS(str(data_store.filepath_processed()), write=True),
        name=clean_datapackage_name(str(data_store.name)),
        sum_intra_duplicates=True,
        sum_inter_duplicates=False,
    )
    dp.add_persistent_vector(
        matrix=data_store.matrix,
        name=clean_datapackage_name(str(data_store.name) + " matrix data"),
        indices_array=indices_array,
        data_array=data_array,
        **extra_metadata
    )
    dp.finalize_serialization()


def dp(fp):
    return load_datapackage(ZipFS(fp))



import copy
import warnings

from bw2data import geomapping
from bw2data.ia_data_store import ImpactAssessmentDataStore

from .meta import intersections
from .validate import intersection_validator
from .utils import create_certain_datapackage


class Intersection(ImpactAssessmentDataStore):
    """ """

    _metadata = intersections
    validator = intersection_validator
    matrix = "intersection_matrix"

    def add_geomappings(self, data):
        """Add all geographic units in both geocollections to ``geomapping``, the master location list.

        Called automatically when data is written."""
        geomapping.add({x[0] for x in data})
        geomapping.add({x[1] for x in data})

    def create_reversed_intersection(self):
        """Create (B, A) intersection from (A, B)."""
        new_name = (self.name[1], self.name[0])
        new_data = [(line[1], line[0], line[2]) for line in self.load()]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            new_obj = Intersection(new_name)
            new_obj.register(**copy.deepcopy(self.metadata))
            new_obj.write(new_data)

        return new_obj

    def process(self, **extra_metadata):
        data = self.load()
        create_certain_datapackage([(geomapping[line[0]], geomapping[line[1]]) for line in data], [line[2] for line in data], self, **extra_metadata)



from bw2data.validate import maybe_uncertainty, valid_tuple
from voluptuous import Any, Invalid, Schema

_maybe_uncertainty = Schema(maybe_uncertainty)
_loading_value = Schema(Any(str, valid_tuple))


def uncertainty_list(obj):
    try:
        assert len(obj) == 2
        assert isinstance(obj, list)
        _maybe_uncertainty(obj[0])
        _loading_value(obj[1])
    except:
        raise Invalid("%s is not a valid loading value" % obj)
    return obj


def float_as_last(obj):
    try:
        assert isinstance(obj, list)
        assert len(obj) == 3
        assert isinstance(obj[2], (float, int))
    except:
        raise Invalid("%s is not a valid intersection value" % obj)
    return obj


def xtable_data(obj):
    try:
        assert isinstance(obj, (tuple, list))
        assert isinstance(obj[0], float)
        assert isinstance(obj[1][0], str)
        assert isinstance(obj[1][1], (str, int))
    except:
        raise Invalid("%s is not a valid xtable value" % obj)
    return obj


loading_validator = Schema([uncertainty_list])
intersection_validator = Schema([float_as_last])
xtable_validator = Schema([xtable_data])



import hashlib
import os
from collections import defaultdict
from functools import partial

import fiona
import numpy as np
from shapely.geometry import shape

from bw2regional import geocollections
from bw2regional.lca.base_class import RegionalizationBase



def _generic_exporter(
    lca,
    geocollection,
    filepath,
    spatial_dict,
    spatial_func,
    score_column_absolute="score_abs",
    score_column_relative="score_rel",
    cutoff=1e-3,
):
    assert isinstance(lca, RegionalizationBase)
    assert geocollection in geocollections
    assert hasattr(lca, spatial_dict)
    assert os.path.isfile(geocollections[geocollection].get("filepath"))
    assert not os.path.isfile(filepath)

    vector = np.ravel(getattr(lca, spatial_func)().sum(axis=0))
    lca.fix_spatial_dictionaries()
    total = lca.score
    cut = abs(lca.score * cutoff)
    field = geocollections[geocollection].get("field")
    assert field

    # TODO: Might need to make this nicer/more robust
    if not filepath.endswith(".geojson"):
        filepath += ".geojson"

    if geocollection == "world":
        # Special case; "world" is just a string, not a tuple
        results = {
            spatial_key: vector[index]
            for spatial_key, index in getattr(lca, spatial_dict).items()
            if abs(vector[index]) >= cut
        }
    else:
        results = {
            spatial_key[1]: vector[index]
            for spatial_key, index in getattr(lca, spatial_dict).items()
            if abs(vector[index]) >= cut
        }

    with fiona.Env():
        with fiona.open(geocollections[geocollection]["filepath"]) as source:
            meta = source.meta
            meta["driver"] = "GeoJSON"
            meta["schema"]["properties"].update(
                {score_column_absolute: "float", score_column_relative: "float"}
            )

            with fiona.open(filepath, "w", **meta) as sink:
                for feature in source:
                    try:
                        score = results[feature["properties"][field]]
                        feature["properties"][score_column_absolute] = score
                        feature["properties"][score_column_relative] = abs(
                            score / total
                        )
                        sink.write(feature)
                    except KeyError:
                        continue


as_inv_spatial_scale = partial(
    _generic_exporter,
    spatial_dict="inv_spatial_dict",
    spatial_func="results_inv_spatial_scale",
)

as_ia_spatial_scale = partial(
    _generic_exporter,
    spatial_dict="ia_spatial_dict",
    spatial_func="results_ia_spatial_scale",
)

as_xt_spatial_scale = partial(
    _generic_exporter,
    spatial_dict="xtable_spatial_dict",
    spatial_func="results_xtable_spatial_scale",
)


def _hash_feature(feature):
    """Calculate SHA256 hash of feature geometry as WKT"""
    geom = shape(feature["geometry"])
    return hashlib.sha256(geom.to_wkt().encode("utf-8")).hexdigest()


def add_two_geojson_results(
    first,
    second,
    output_filepath,
    first_column_name="score_abs",
    second_column_name="score_abs",
    score_column_absolute="score_abs",
    score_column_relative="score_rel",
    cutoff=1e-4,
):
    """Sum results from two regionalized LCA calculations on the same spatial scale."""
    results, features = defaultdict(float), {}

    # TODO: Might need to make this nicer/more robust
    if not output_filepath.endswith(".geojson"):
        output_filepath += ".geojson"

    with fiona.open(first, "r") as obj:
        meta = obj.meta

        for feature in obj:
            hashed = _hash_feature(feature)
            features[hashed] = feature
            results[hashed] += feature["properties"][first_column_name]

    with fiona.open(second, "r") as obj:
        for feature in obj:
            hashed = _hash_feature(feature)
            features[hashed] = feature
            results[hashed] += feature["properties"][second_column_name]

    total = sum(results.values())

    meta["driver"] = "GeoJSON"
    meta["schema"]["properties"].update(
        {score_column_absolute: "float", score_column_relative: "float"}
    )

    with fiona.open(output_filepath, "w", **meta) as sink:
        for k, feature in features.items():
            score = results[k]
            if abs(score) < abs(total * cutoff):
                continue

            feature["properties"][score_column_absolute] = score
            feature["properties"][score_column_relative] = abs(score / total)
            sink.write(feature)



from bw2data import methods

from ..errors import GeocollectionsMismatch
from .base_class import RegionalizationBase


class OneSpatialScaleLCA(RegionalizationBase):
    def __init__(self, *args, **kwargs):
        r"""Perform regionalized LCA calculation, where the inventory shares the same spatial scale as impact assessment.

        The calculation formula is:

        .. math::

            h_{r} = \left[ \textbf{MR} \right]^{T} \circ [ \textbf{B} \cdot (\textbf{A}^{-1}f) ]

        Uses sparse matrix `elementwise multiplication <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.multiply.html>`_.

        """
        super(OneSpatialScaleLCA, self).__init__(*args, **kwargs)
        if self.method not in methods:
            raise ValueError("Must pass valid `method` name")
        self.inventory_geocollections = self.get_inventory_geocollections()
        self.ia_geocollections = self.get_ia_geocollections()
        missing = self.ia_geocollections.difference(self.inventory_geocollections)
        if missing:
            raise GeocollectionsMismatch(
                "{} geocollection(s) needed by IA method but not in inventory".format(
                    missing
                )
            )

    def load_lcia_data(self):
        self.create_inventory_mapping_matrix()
        self.create_regionalized_characterization_matrix(self.inv_mapping_mm.col_mapper)

    def lcia_calculation(self):
        """Do regionalized LCA calculation.

        Creates ``self.characterized_inventory``.

        """
        self.characterized_inventory = (
            self.inv_mapping_matrix * self.reg_cf_matrix
        ).T.multiply(self.inventory)

    def results_ia_spatial_scale(self):
        raise NotImplementedError("No separate IA spatial scale")

    def results_inv_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return self.reg_cf_matrix.T.multiply(self.inventory * self.inv_mapping_matrix)



import numpy as np
import rasterio


def get_area(lat1, lat2, width):
    """Get area of a spherical quadrangle.

    lat1, lat2, and width should all be in degrees.

    Uses the formula derived and demonstrated in https://gis.stackexchange.com/questions/127165/more-accurate-way-to-calculate-area-of-rasters."""
    a = 6378137  # Meters
    b = 6356752.3142  # Meters
    to_radians = lambda x: x / 180 * np.pi

    e = 0.08181919084296  # e = sqrt(1 - (b/a)^2)

    width /= 360  # No wrap around from way rasters are defined

    def area(latitude):
        o = np.sin(to_radians(latitude))
        return (
            np.pi
            * b ** 2
            * (2 * np.arctanh(e * o) / (2 * e) + o / ((1 + e * o) * (1 - e * o)))
        )

    if lat1 >= 0 and lat2 <= 0:
        return width * (area(lat1) + area(lat2))
    else:
        return width * abs(area(lat1) - area(lat2))


def get_column_array(affine, rows, width):
    return np.array(
        [
            get_area(affine[5] + i * affine[4], affine[5] + (i + 1) * affine[4], width)
            for i in range(rows)
        ]
    ).reshape((1, -1, 1))


def divide_by_area(source_fp, destination_fp):
    """Create a new raster file at ``destination_fp``, dividing the values in ``source_fp`` by their cell's area.

    Will raise an error is the CRS is not geographic, or the raster is rotated."""
    with rasterio.open(source_fp) as src:
        meta = src.meta
        affine = src.transform
        rotated = affine[1] or affine[3]
        projected = "WGS 84" not in src.meta["crs"].wkt
        if rotated or projected:
            ERROR = """This function can't process projected or rotated rasters.
            Try https://geographiclib.sourceforge.io/html/python/interface.html"""
            raise ValueError(ERROR)

        areas = get_column_array(affine, src.height, abs(affine[0]))
        original = src.read()
        array = original / areas.astype(original.dtype)
        array[original == meta["nodata"]] = meta["nodata"]

        with rasterio.open(destination_fp, "w", **src.meta) as sink:
            for i, a in enumerate(array):
                sink.write(a, i + 1)



import itertools
from functools import partial

import numpy as np
from bw2calc.lca import LCA
from bw2data import Database, Method, databases, methods
from scipy.sparse import csr_matrix
import matrix_utils as mu

from ..errors import MissingIntersection, SiteGenericMethod, UnprocessedDatabase
from ..intersection import Intersection
from ..meta import intersections
from ..utils import dp


class RegionalizationBase(LCA):
    def __init__(self, demand, *args, **kwargs):
        super(RegionalizationBase, self).__init__(demand, *args, **kwargs)
        self.databases = set.union(
            *[Database(key[0]).find_graph_dependents() for key in demand]
        )

    def get_inventory_geocollections(self):
        """Get the set of all needed inventory geocollections.

        Raise UnprocessedDatabase if any database is missing the required metadata."""
        missing, present = [], set()
        for database in self.databases:
            if "geocollections" not in databases[database]:
                missing.append(database)
            else:
                present.update(set(databases[database]["geocollections"]))
        if missing:
            raise UnprocessedDatabase(
                "Database(s) {} don't specify their geocollections.".format(missing)
            )
        return present

    def get_ia_geocollections(self):
        """Retrieve the geocollections linked to the impact assessment method"""
        try:
            return set(methods[self.method]["geocollections"])
        except KeyError:
            raise SiteGenericMethod

    def create_inventory_mapping_matrix(self):
        """Get inventory mapping matrix, **M**, which maps inventory activities to inventory locations. Rows are inventory activities and columns are inventory spatial units.

        Uses ``self.technosphere_mm.row_mapper`` and ``self.databases``.

        Creates ``self.inv_mapping_mm``, ``self.inv_mapping_matrix``, and ``self.dicts.inv_spatial``/

        """
        self.inv_mapping_mm = mu.MappedMatrix(
            packages=[dp(Database(x).filepath_processed()) for x in self.databases],
            matrix="inv_geomapping_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            row_mapper=self.technosphere_mm.col_mapper,
        )
        self.inv_mapping_matrix = self.inv_mapping_mm.matrix
        self.dicts.inv_spatial = partial(self.inv_mapping_mm.col_mapper.to_dict)

    def needed_intersections(self):
        """Figure out which ``Intersection`` objects are needed bsed on ``self.inventory_geocollections`` and ``self.ia_geocollections``.

        Raise ``MissingIntersection`` if an intersection is required, but not available."""
        required = list(
            itertools.product(self.inventory_geocollections, self.ia_geocollections)
        )
        for obj in required:
            if obj not in intersections:
                raise MissingIntersection(
                    "Intersection {} needed but not found".format(obj)
                )
        return required

    def create_geo_transform_matrix(self):
        """Get geographic transform matrix **G**, which gives the intersecting areas of inventory and impact assessment spatial units. Rows are inventory spatial units, and columns are impact assessment spatial units.

        Uses ``self.inv_spatial_dict`` and ``self.ia_spatial_dict``.

        Returns:
            * ``geo_transform_params``: Parameter array with row/col of inventory and IA locations
            * ``geo_transform_matrix``: The matrix **G**

        """
        self.geo_transform_mm = mu.MappedMatrix(
            packages=[
                dp(Intersection(name).filepath_processed())
                for name in self.needed_intersections()
            ],
            matrix="intersection_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            col_mapper=self.reg_cf_mm.row_mapper,
            row_mapper=self.inv_mapping_mm.col_mapper,
        )
        self.geo_transform_matrix = self.geo_transform_mm.matrix

    def create_regionalized_characterization_matrix(self, row_mapper=None):
        """Get regionalized characterization matrix, **R**, which gives location- and biosphere flow-specific characterization factors.

        Rows are impact assessment spatial units, and columns are biosphere flows. However, we build it transverse and transpose it, as the characterization matrix indices are provided that way.

        Uses ``self._biosphere_dict`` and ``self.method``.

        Returns:
            * ``reg_cf_params``: Parameter array with row/col of IA locations/biosphere flows
            * ``ia_spatial_dict``: Dictionary linking impact assessment locations to matrix rows
            * ``reg_cf_matrix``: The matrix **R**

        """
        self.reg_cf_mm = mu.MappedMatrix(
            packages=[dp(Method(self.method).filepath_processed())],
            matrix="characterization_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            col_mapper=self.biosphere_mm.row_mapper,
            row_mapper=row_mapper,
            transpose=True,
        )
        self.reg_cf_matrix = self.reg_cf_mm.matrix
        if row_mapper is None:
            self.dicts.ia_spatial = partial(self.reg_cf_mm.row_mapper.to_dict)

    def create_loading_matrix(self):
        """Get diagonal regionalized loading matrix, **L**, which gives location-specific background loading factors. Dimensions are impact assessment spatial units.

        Uses ``self.dicts.ia_spatial``.

        """
        self.loading_mm = mu.MappedMatrix(
            packages=[dp(self.loading.filepath_processed())],
            matrix="loading_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            diagonal=True,
            row_mapper=self.reg_cf_mm.row_mapper,
        )
        self.loading_matrix = self.loading_mm.matrix

    # def fix_spatial_dictionaries(self):
    #     """Fix inventory and IA spatial dictionaries."""
    #     if not hasattr(self, "inv_spatial_dict"):
    #         # No LCIA performed yet
    #         return
    #     if getattr(self, "_mapped_spatial_dict", False):
    #         # Already reversed - should be idempotent
    #         return False
    #     rev_geomapping = {v: k for k, v in geomapping.items()}
    #     self.inv_spatial_dict = {
    #         rev_geomapping[k]: v for k, v in self.inv_spatial_dict.items()
    #     }
    #     if hasattr(self, "ia_spatial_dict"):
    #         self.ia_spatial_dict = {
    #             rev_geomapping[k]: v for k, v in self.ia_spatial_dict.items()
    #         }
    #     if hasattr(self, "xtable_spatial_dict"):
    #         self.xtable_spatial_dict = {
    #             rev_geomapping[k]: v for k, v in self.xtable_spatial_dict.items()
    #         }
    #     self._mapped_spatial_dict = True

    def _results_new_scale(self, matrix, flow):
        # self.fix_spatial_dictionaries()
        if flow is not None:
            try:
                row_index = self.biosphere_dict[flow]
                matrix = matrix[row_index, :]
            except KeyError:
                raise ValueError("Flow {} not in biosphere dict".format(flow))
        else:
            # using matrix.sum() converts to dense numpy matrix
            nc = matrix.shape[0]
            summer = csr_matrix(
                (np.ones(nc), np.arange(nc), np.array((0, nc), dtype=int))
            )
            matrix = summer * matrix
        return matrix

    def results_ia_spatial_scale(self):
        raise NotImplementedError("Must be defined in subclasses")

    def results_inv_spatial_scale(self):
        raise NotImplementedError("Must be defined in subclasses")



class BW2RegionalizationError(Exception):
    """Base class for BW2 regionalization errors"""

    pass


class UnprocessedDatabase(BW2RegionalizationError):
    """A ``Database`` object doesn't have a list of reference geocollections."""

    pass


class SiteGenericMethod(BW2RegionalizationError):
    """This ``Method`` doesn't have links to ``geocollections``, making it site-generic."""

    pass


class MissingIntersection(BW2RegionalizationError):
    """Missing an ``Intersection`` object and its data needed for regionalized LCA"""

    pass


class GeocollectionsMismatch(BW2RegionalizationError):
    pass


class MissingSpatialSourceData(BW2RegionalizationError):
    pass


class TopologyError(BW2RegionalizationError):
    """Inventory includes locations for which no topology data is available"""

    pass


class IncompleteSpatialDefinition(BW2RegionalizationError):
    """Given metadata is not enough to understand a spatial data source"""

    pass


class WindowsPathCharacterLimit(BW2RegionalizationError):
    """Windows has an absolute limit of 255 characters in a filepath"""

    pass



# import numpy as np
# from bw2data import Database
# from scipy.sparse import diags

# from .extension_tables import ExtensionTablesLCA


# class TopologicalExtensionTablesLCA(ExtensionTablesLCA):
#     """LCA class that handles intersections that work directly with topological face ids.

#     This code is not recommended - please use the topological merging functions instead.

#     """

#     def __init__(self, *args, **kwargs):
#         super(TopologicalExtensionTablesLCA, self).__init__(*args, **kwargs)

#     def needed_inv_xtable_intersections(self):
#         return [("ecoinvent-topology", obj) for obj in self.xtable_geocollections]

#     def get_inventory_mapping_matrix(self, builder=MatrixBuilder):
#         inv_mapping_params, _, inv_spatial_dict, inv_mapping_matrix = builder.build(
#             paths=[Database(x).filepath_geomapping() for x in self.databases],
#             data_label="amount",
#             row_id_label="activity",
#             row_index_label="row",
#             col_id_label="geo",
#             col_index_label="col",
#             row_dict=self._activity_dict,
#         )
#         return (inv_mapping_params, inv_spatial_dict, inv_mapping_matrix)

#     def build_topo_normalization_matrix(self):
#         r"""Get normalization matrix for inventory-xtable mapping. Normalizes to

#         .. math::
#             ( \textbf{N}_{t} )_{i, i} = \left[ \sum_{j} \textbf{MDX}_{i, j} \right]^{-1}

#         """
#         vector = np.array(
#             (
#                 self.inv_mapping_matrix * self.distribution_matrix * self.xtable_matrix
#             ).sum(axis=1)
#         ).T
#         mask = vector > 0
#         vector[mask] = 1 / vector[mask]
#         return diags(vector, [0], format="csr", dtype=np.float32)

#     def load_lcia_data(self, builder=MatrixBuilder):
#         (
#             self.inv_mapping_params,
#             self.inv_spatial_dict,
#             self.inv_mapping_matrix,
#         ) = self.get_inventory_mapping_matrix()
#         (
#             self.distribution_params,
#             self.xtable_spatial_dict,
#             self.distribution_matrix,
#         ) = self.build_distribution_matrix()
#         self.xtable_params, self.xtable_matrix = self.load_xtable_matrix()
#         self.topo_normalization_matrix = self.build_topo_normalization_matrix()
#         (
#             self.reg_cf_params,
#             self.ia_spatial_dict,
#             self.reg_cf_matrix,
#         ) = self.get_regionalized_characterization_matrix(builder)
#         (
#             self.geo_transform_params,
#             self.geo_transform_matrix,
#         ) = self.get_geo_transform_matrix(builder)
#         self.geo_transform_normalization_matrix = (
#             self.build_geo_transform_normalization_matrix()
#         )

#     def lcia_calculation(self):
#         """Do regionalized LCA calculation.

#         Creates ``self.characterized_inventory``.

#         """
#         self.characterized_inventory = (
#             self.topo_normalization_matrix
#             * self.inv_mapping_matrix
#             * self.distribution_matrix
#             * self.xtable_matrix
#             * self.geo_transform_normalization_matrix
#             * self.geo_transform_matrix
#             * self.reg_cf_matrix
#         ).T.multiply(self.inventory)

#     def results_ia_spatial_scale(self):
#         if not hasattr(self, "characterized_inventory"):
#             raise ValueError("Must do lcia calculation first")
#         return self.reg_cf_matrix.T.multiply(
#             self.inventory
#             * self.topo_normalization_matrix
#             * self.inv_mapping_matrix
#             * self.distribution_matrix
#             * self.xtable_matrix
#             * self.geo_transform_normalization_matrix
#             * self.geo_transform_matrix
#         )

#     def results_inv_spatial_scale(self):
#         raise NotImplementedError

#     def results_xtable_spatial_scale(self):
#         if not hasattr(self, "characterized_inventory"):
#             raise ValueError("Must do lcia calculation first")
#         return (
#             self.geo_transform_normalization_matrix
#             * self.geo_transform_matrix
#             * self.reg_cf_matrix
#         ).T.multiply(
#             self.inventory
#             * self.topo_normalization_matrix
#             * self.inv_mapping_matrix
#             * self.distribution_matrix
#             * self.xtable_matrix
#         )



import warnings

import pyprind
from bw2data import Database, databases

from . import Topography, geocollections, topocollections
from .base_data import COUNTRIES


def label_activity_geocollections(name):
    """Add geocollections to activity ``location`` fields.

    ``name`` is the name of an existing LCI database."""
    assert name in databases, "{} not found".format(name)
    assert (
        "world" in geocollections
    ), "Please run `create_world_collections` first"

    ecoinvent = (
        {x[1] for x in Topography("ecoinvent").load()}
        if "ecoinvent" in topocollections
        else set()
    )
    RoWs = (
        {x[1] for x in Topography("RoW").load()} if "RoW" in topocollections else set()
    )

    db = Database(name)
    searchable = db.metadata.get("searchable")
    if searchable:
        db.make_unsearchable()

    found_geocollections = set()

    locations = {x["location"] for x in db}
    assert "RoW" not in locations, "`RoW` found; use `rower` to label Rest-of-Worlds"

    for act in pyprind.prog_bar(db):
        if isinstance(act["location"], tuple):
            found_geocollections.add(act["location"][0])
        elif act["location"] in COUNTRIES:
            found_geocollections.add("world")
        elif act["location"] == "GLO":
            found_geocollections.add("world")
        elif act["location"] in RoWs:
            act["location"] = ("RoW", act["location"])
            act.save()
            found_geocollections.add("RoW")
        elif act["location"] in ecoinvent:
            act["location"] = ("ecoinvent", act["location"])
            act.save()
            found_geocollections.add("ecoinvent")
        else:
            warnings.warn(
                (
                    "Location {} in {} not understood; please add geocollection"
                    " manually, and add to databases[name]['geocollections']"
                ).format(act["location"], act.key)
            )

    if searchable:
        db.make_searchable()
    db.process()

    db.metadata["regionalized"] = True
    db.metadata["geocollections"] = sorted(found_geocollections)
    databases.flush()



from setuptools import setup
import os


requirements = [
    "bw2calc>=2.0.dev2",
    "bw2data>=4.0.dev7",
    "constructive_geometries",
    "fiona",
    "pandas",
    "pandarus>=1.0",
    "rasterio",
    "requests",
    "shapely",
    "voluptuous",
    "wrapt",
]

v_temp = {}
with open("bw2regional/version.py") as fp:
    exec(fp.read(), v_temp)
version = ".".join((str(x) for x in v_temp["version"]))


setup(
    name="bw2regional",
    version=version,
    packages=["bw2regional", "bw2regional.lca"],
    author="Chris Mutel",
    author_email="cmutel@gmail.com",
    license="NewBSD 3-clause; LICENSE.txt",
    url="https://github.com/brightway-lca/brightway2-regional",
    install_requires=[] if os.environ.get("READTHEDOCS") else requirements,
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



import itertools

import numpy as np
from scipy.sparse import diags
import matrix_utils as mu
from functools import partial

from ..errors import MissingIntersection
from ..intersection import Intersection
from ..meta import extension_tables, intersections
from ..utils import filter_columns, filter_rows, dp
from ..xtables import ExtensionTable
from .base_class import RegionalizationBase


class ExtensionTablesLCA(RegionalizationBase):
    def __init__(self, *args, **kwargs):
        r"""Perform regionalized LCA calculation, using biosphere flow- and activity-specific extension tables.

        The calculation formula is:

        .. math::

            h_{r} = \left[ \textbf{MN}_{dx} \textbf{DXN}_{g} \textbf{GR} \right]^{T} \circ [ \textbf{B} \cdot (\textbf{A}^{-1}f) ]

        Extension tables provide detailed information on the spatial pattern of occurrence for one, some, or all  activities, and one, some, or all of their environmental flows. In contrast with background loadings, which are on the spatial scale of impact assessment, extension tables provide data on a new spatial scale which is linked to inventory activities. This third spatial scale (inventory, impact assessment, and extension table) makes the calculation formula more complicated, and makes two normalization steps necessary.

        **Foo**

        **X** is the extension table mapping matrix, a diagonal matrix that indicates the relative importance of each extension table spatial unit. It has dimensions of *extension spatial scale* by *extension spatial scale*.

        :math:`N_{x}` is the extension

        **L** is the extension table

        The idea of extension tables is rather flexible, and this class supports limiting activities and flows with the ``limitations`` object:

        .. code-block:: python

            limitations = {
                'activities': list of activities to include/exclude,
                'activities mode': 'exclude',  # or 'include',
                'flows': list of flows to include/exclude,
                'flows mode': 'exclude',  # or 'include'

            }

        The ``limitations`` object can have activities, flows, or both. An ``exclude`` mode means that these activities or flows are removed from the regionalized matrices; an ``include`` mode means that only the specified activities and flows are kept. The default mode is ``include``.

        """
        try:
            xtable = kwargs.pop("xtable")
            limitations = kwargs.pop("limitations", {})
        except KeyError:
            raise ValueError("``xtable`` kwarg required")
        assert xtable in extension_tables
        super(ExtensionTablesLCA, self).__init__(*args, **kwargs)
        self.xtable = ExtensionTable(xtable)
        self.xtable_meta = extension_tables[xtable]
        self.limitations = limitations

        if "activities mode" in self.limitations:
            if self.limitations["activities mode"] not in ("exclude, include"):
                raise ValueError(
                    "`activities mode` must be either `exclude` or `include`."
                )
        if "flows mode" in self.limitations:
            if self.limitations["flows mode"] not in ("exclude, include"):
                raise ValueError("`flows mode` must be either `exclude` or `include`.")

        self.inventory_geocollections = self.get_inventory_geocollections()
        self.ia_geocollections = self.get_ia_geocollections()
        (
            self.xtable_geocollections,
            self.xtable_intersections,
        ) = self.get_xtable_geodata()

        self.inv_xtable_intersections = self.needed_inv_xtable_intersections()
        self.xtable_ia_intersections = self.needed_xtable_ia_intersections()
        self.check_geocollection_intersections()

    def needed_inv_xtable_intersections(self):
        return list(
            itertools.product(self.inventory_geocollections, self.xtable_geocollections)
        )

    def needed_xtable_ia_intersections(self):
        return list(
            itertools.product(self.xtable_geocollections, self.ia_geocollections)
        )

    def check_geocollection_intersections(self):
        missing_inv_xtable_inters = [
            x for x in self.inv_xtable_intersections if x not in intersections
        ]
        missing_xtable_ia_inters = [
            x for x in self.xtable_ia_intersections if x not in intersections
        ]
        if missing_inv_xtable_inters or missing_xtable_ia_inters:
            error_text = """
Some Intersection objects are required for this calculation, but were not found.

If you know these intersections are not needed, you can create empty intersections as follows:

    from bw2regional import create_empty_intersection
    create_empty_intersection(('missing', 'name'))

            """
            if missing_inv_xtable_inters:
                error_text += "Intersections between inventory and extension tables:"
                error_text += "\n\t".join([str(x) for x in missing_inv_xtable_inters])
                error_text += "\n"
            if missing_xtable_ia_inters:
                error_text += "Intersections between extension tables and IA method:"
                error_text += "\n\t".join([str(x) for x in missing_xtable_ia_inters])
            raise MissingIntersection(error_text)

    def get_xtable_geodata(self):
        gc = self.xtable_meta["geocollection"]
        return (
            [gc],  # Only one geocollection per extension table, but need list
            {x for x in intersections if x[1] == gc},
        )

    def create_distribution_matrix(self):
        """Get distribution matrix, **D**, which provides the area of inventory spatial units in each extension table spatial unit. Rows are inventory spatial units and columns are extension table spatial units."""
        self.distribution_mm = mu.MappedMatrix(
            packages=[
                dp(Intersection(name).filepath_processed())
                for name in self.inv_xtable_intersections
            ],
            matrix="intersection_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            row_mapper=self.inv_mapping_mm.col_mapper,
        )
        self.distribution_matrix = self.distribution_mm.matrix
        self.dicts.xtable_spatial = partial(self.distribution_mm.col_mapper.to_dict)

    def create_xtable_matrix(self):
        """Diagonal extension table matrix that indicates the extension table density value in each extension table spatial unit."""
        self.xtable_mm = mu.MappedMatrix(
            packages=[dp(self.xtable.filepath_processed())],
            matrix="xtable_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            diagonal=True,
            row_mapper=self.distribution_mm.col_mapper,
        )
        self.xtable_matrix = self.xtable_mm.matrix

    def build_distribution_normalization_matrix(self):
        r"""Get normalization matrix for inventory-xtable mapping. Normalizes to

        .. math::
            ( \textbf{N}_{dx} )_{i, i} = \left[ \sum_{j} \textbf{DX}_{i, j} \right]^{-1}

        """
        vector = np.array((self.distribution_matrix * self.xtable_matrix).sum(axis=1)).T
        mask = vector > 0
        vector[mask] = 1 / vector[mask]
        return diags(vector, [0], format="csr", dtype=np.float32)

    def build_geo_transform_normalization_matrix(self):
        r"""Get normalization value for areas in each IA spatial unit:

        .. math::
            ( \textbf{N}_{g} )_{i, i} = \left[ \sum_{j} \textbf{G}_{i, j} \right]^{-1}

        """
        vector = np.array(self.geo_transform_matrix.sum(axis=1)).T
        mask = vector > 0
        vector[mask] = 1 / vector[mask]
        return diags(vector, [0], format="csr", dtype=np.float32)

    def create_geo_transform_matrix(self):
        self.geo_transform_mm = mu.MappedMatrix(
            packages=[
                dp(Intersection(name).filepath_processed())
                for name in self.xtable_ia_intersections
            ],
            matrix="intersection_matrix",
            use_arrays=self.use_arrays,
            use_distributions=self.use_distributions,
            seed_override=self.seed_override,
            col_mapper=self.reg_cf_matrix.row_mapper,
            row_mapper=self.distribution_mm.col_mapper,
        )
        self.geo_transform_matrix = self.geo_transform_mm.matrix

    def load_lcia_data(self):
        self.create_inventory_mapping_matrix()

        if "activities" in self.limitations:
            self.inv_mapping_matrix = self.inv_mapping_mm.matrix = filter_rows(
                self.inv_mapping_matrix,
                [self.dicts.activity[x] for x in self.limitations["activities"]],
                exclude=self.limitations.get("activities mode", None) == "exclude",
            )

        self.create_distribution_matrix()
        self.create_xtable_matrix()
        self.distribution_normalization_matrix = (
            self.build_distribution_normalization_matrix()
        )
        self.create_regionalized_characterization_matrix()

        if "flows" in self.limitations:
            self.reg_cf_matrix = self.reg_cf_mm.matrix = filter_columns(
                self.reg_cf_matrix,
                [self.dicts.biosphere[x] for x in self.limitations["flows"]],
                exclude=self.limitations.get("flows mode", None) == "exclude",
            )

        self.create_geo_transform_matrix()
        self.geo_transform_normalization_matrix = (
            self.build_geo_transform_normalization_matrix()
        )

    def lcia_calculation(self):
        """Do regionalized LCA calculation.

        Creates ``self.characterized_inventory``.

        """
        self.characterized_inventory = (
            self.inv_mapping_matrix
            * self.distribution_normalization_matrix
            * self.distribution_matrix
            * self.xtable_matrix
            * self.geo_transform_normalization_matrix
            * self.geo_transform_matrix
            * self.reg_cf_matrix
        ).T.multiply(self.inventory)

    def results_ia_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return self.reg_cf_matrix.T.multiply(
            self.inventory
            * self.inv_mapping_matrix
            * self.distribution_normalization_matrix
            * self.distribution_matrix
            * self.xtable_matrix
            * self.geo_transform_normalization_matrix
            * self.geo_transform_matrix
        )

    def results_inv_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return (
            self.distribution_normalization_matrix
            * self.distribution_matrix
            * self.xtable_matrix
            * self.geo_transform_normalization_matrix
            * self.geo_transform_matrix
            * self.reg_cf_matrix
        ).T.multiply(self.inventory * self.inv_mapping_matrix)

    def results_xtable_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return (
            self.geo_transform_normalization_matrix
            * self.geo_transform_matrix
            * self.reg_cf_matrix
        ).T.multiply(
            self.inventory
            * self.inv_mapping_matrix
            * self.distribution_normalization_matrix
            * self.distribution_matrix
            * self.xtable_matrix
        )



from bw2data import DataStore, geomapping

from .meta import topocollections


class Topography(DataStore):
    """A topographical description of a ``geocollection``.

    A topography must be registered to exactly one geocollection.

    For example, the geocollection 'countries' could have the location 'Ghana', and include a spatial data file which defines 'Ghana'. The `Topography` 'countries-topo' would have two files:

        #. A mapping file that linked each location in the geocollection 'countries' to a set of topographical faces
        #. A spatial dataset which defines the topographical faces

    Topographies can make some calculations much quicker for two reasons:

        #. Large features are split into smaller faces, making GIS calculations easier and spatial index queries more efficient, as bounding boxes are smaller
        #. You only do GIS work on each topographical face once

    The LCA classes in ``bw2regional`` don't work directly with the `Topography` objects; rather, GIS calculations on topographies can be merged to the spatial features of the `geocollection`.

    The usual order of operations with a `geocollection` that has a `Topography` is the following:

        #. Create a `geocollection`, including defining a spatial dataset
        #. Create a new `topocollection` and specify the linked `geocollection`. You can also optionally define another spatial dataset, using the same format as for `geocollections`.
        #. Instantiate the new `Topography` object created in the earlier step, and write mapping data from spatial features in the `geocollection`(s) to face ids in the `Topography` spatial data set.

    The data format for mapping data is ``{feature field value: [list of topo field values (usually id numbers)]}``.

    Here is a code sample for using the test data in `bw2regional`:

    .. code-block:: python

        from bw2regional import Topography, geocollections
        from bw2regional.tests import data_dir
        import json

        geocollections['countries'] = {
            'filepath': os.path.join(data_dir, "test_countries.gpkg"),
            'field': 'name'
        }
        topocollections['countries-topo'] = {
            'geocollection': 'countries',
            'filepath': os.path.join(data_dir, "test_provinces.gpkg"),
            'field': 'OBJECTID_1'
        }
        topo = Topography('countries-topo')
        topo.write(json.load(open(os.path.join(data_dir, "test_topo_mapping.json"))))

    """

    _metadata = topocollections

    @property
    def geocollection(self):
        return self.metadata["geocollection"]

    def add_geomappings(self, data):
        geomapping.add(data)

    def write(self, data):
        self.metadata["empty"] = False
        self._metadata.flush()
        self.add_geomappings(data)
        super(Topography, self).write(data)



# -*- coding: utf-8 -*-
from .extension_tables import ExtensionTablesLCA
from .one_spatial_scale import OneSpatialScaleLCA
from .two_spatial_scales import TwoSpatialScalesLCA
from .two_spatial_scales_weighting import TwoSpatialScalesWithGenericLoadingLCA



import numpy as np
from bw2data import methods
from scipy.sparse import diags

from .base_class import RegionalizationBase


class TwoSpatialScalesLCA(RegionalizationBase):
    def __init__(self, *args, **kwargs):
        r"""Perform regionalized LCA calculation, matching the spatial scales of inventory and impact assessment.

        The calculation formula is:

        .. math::

            h_{r} = \left[ \textbf{MNGR} \right]^{T} \circ [ \textbf{B} \cdot (\textbf{A}^{-1}f) ]

        Uses sparse matrix `elementwise multiplication <http://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.multiply.html>`_.

        In addition to the normal LCA initialization steps, this class does the following:

            * Make sure that each inventory database has a set of ``geocollections`` in its metadata.

        """
        super(TwoSpatialScalesLCA, self).__init__(*args, **kwargs)
        if self.method not in methods:
            raise ValueError("Must pass valid `method` name")
        self.inventory_geocollections = self.get_inventory_geocollections()
        self.ia_geocollections = self.get_ia_geocollections()

    def load_lcia_data(self):
        self.create_inventory_mapping_matrix()
        self.create_regionalized_characterization_matrix()
        self.create_geo_transform_matrix()
        self.normalization_matrix = self.build_normalization_matrix()

    def build_normalization_matrix(self):
        r"""Get normalization matrix, a diagonal matrix.

        .. math::
            \textbf{N}_{i,i} = \left[ \sum_{j} \left( \textbf{G} \right)_{i,j} \right]^{-1}

        """
        vector = np.array(self.geo_transform_matrix.sum(axis=1)).T
        mask = vector > 0
        vector[mask] = 1 / vector[mask]
        return diags(vector, [0], format="csr", dtype=np.float32)

    def lcia_calculation(self):
        """Do regionalized LCA calculation.

        Creates ``self.characterized_inventory``.

        """
        self.characterized_inventory = (
            self.inv_mapping_matrix
            * self.normalization_matrix
            * self.geo_transform_matrix
            * self.reg_cf_matrix
        ).T.multiply(self.inventory)

    def results_ia_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return self.reg_cf_matrix.T.multiply(
            self.inventory
            * self.inv_mapping_matrix
            * self.normalization_matrix
            * self.geo_transform_matrix
        )

    def results_inv_spatial_scale(self):
        if not hasattr(self, "characterized_inventory"):
            raise ValueError("Must do lcia calculation first")
        return (
            self.normalization_matrix * self.geo_transform_matrix * self.reg_cf_matrix
        ).T.multiply(self.inventory * self.inv_mapping_matrix)
